@@ -1,9 +1,17 @@
-from parcels import FieldSet, ParticleSet, ScipyParticle, JITParticle, AdvectionRK4, ParticleFile
 from argparse import ArgumentParser
+from datetime import timedelta as delta
+from glob import glob
+from os import path
+
 import numpy as np
 import pytest
-from datetime import timedelta as delta
-from os import path
+
+from parcels import AdvectionRK4
+from parcels import FieldSet
+from parcels import JITParticle
+from parcels import ParticleFile
+from parcels import ParticleSet
+from parcels import ScipyParticle
 
 ptype = {'scipy': ScipyParticle, 'jit': JITParticle}
 
@@ -27,7 +35,7 @@ def run_nemo_curvilinear(mode, outfile):
     lonp = 30 * np.ones(npart)
     latp = [i for i in np.linspace(-70, 88, npart)]
 
-    def periodicBC(particle, pieldSet, time, dt):
+    def periodicBC(particle, fieldSet, time):
         if particle.lon > 180:
             particle.lon -= 360
 
@@ -63,11 +71,34 @@ def make_plot(trajfile):
 
 
 @pytest.mark.parametrize('mode', ['jit'])  # Only testing jit as scipy is very slow
-def test_nemo_curvilinear(mode):
-    outfile = 'nemo_particles'
+def test_nemo_curvilinear(mode, tmpdir):
+    outfile = tmpdir.join('nemo_particles')
     run_nemo_curvilinear(mode, outfile)
 
 
+def test_nemo_3D_samegrid():
+    data_path = path.join(path.dirname(__file__), 'NemoNorthSeaORCA025-N006_data/')
+    ufiles = sorted(glob(data_path + 'ORCA*U.nc'))
+    vfiles = sorted(glob(data_path + 'ORCA*V.nc'))
+    wfiles = sorted(glob(data_path + 'ORCA*W.nc'))
+    mesh_mask = data_path + 'coordinates.nc'
+
+    filenames = {'U': {'lon': mesh_mask, 'lat': mesh_mask, 'depth': wfiles[0], 'data': ufiles},
+                 'V': {'lon': mesh_mask, 'lat': mesh_mask, 'depth': wfiles[0], 'data': vfiles},
+                 'W': {'lon': mesh_mask, 'lat': mesh_mask, 'depth': wfiles[0], 'data': wfiles}}
+
+    variables = {'U': 'uo',
+                 'V': 'vo',
+                 'W': 'wo'}
+    dimensions = {'U': {'lon': 'glamf', 'lat': 'gphif', 'depth': 'depthw', 'time': 'time_counter'},
+                  'V': {'lon': 'glamf', 'lat': 'gphif', 'depth': 'depthw', 'time': 'time_counter'},
+                  'W': {'lon': 'glamf', 'lat': 'gphif', 'depth': 'depthw', 'time': 'time_counter'}}
+
+    fieldset = FieldSet.from_nemo(filenames, variables, dimensions)
+
+    assert fieldset.U.dataFiles is not fieldset.W.dataFiles
+
+
 if __name__ == "__main__":
     p = ArgumentParser(description="""Chose the mode using mode option""")
     p.add_argument('--mode', choices=('scipy', 'jit'), nargs='?', default='jit',
