@@ -1,17 +1,26 @@
-from parcels.kernel import Kernel
-from parcels.particle import JITParticle
-from parcels.compiler import GNUCompiler
-from parcels.kernels.advection import AdvectionRK4
-from parcels.particlefile import ParticleFile
-from parcels.tools.loggers import logger
-from parcels.grid import GridCode
-from parcels.field import NestedField, SummedField
+import collections
+import time as time_module
+from datetime import date
+from datetime import datetime
+from datetime import timedelta as delta
+
 import numpy as np
 import progressbar
-import time as time_module
-import collections
-from datetime import timedelta as delta
-from datetime import datetime, date
+
+from parcels.compiler import GNUCompiler
+from parcels.field import NestedField
+from parcels.field import SummedField
+from parcels.grid import GridCode
+from parcels.kernel import Kernel
+from parcels.kernels.advection import AdvectionRK4
+from parcels.particle import JITParticle
+from parcels.particlefile import ParticleFile
+from parcels.tools.loggers import logger
+try:
+    from mpi4py import MPI
+    from sklearn.cluster import KMeans
+except:
+    MPI = None
 
 __all__ = ['ParticleSet']
 
@@ -32,49 +41,87 @@ class ParticleSet(object):
     :param lonlatdepth_dtype: Floating precision for lon, lat, depth particle coordinates.
            It is either np.float32 or np.float64. Default is np.float32 if fieldset.U.interp_method is 'linear'
            and np.float64 if the interpolation method is 'cgrid_velocity'
+    :param partitions: List of cores on which to distribute the particles for MPI runs. Default: None, in which case particles
+           are distributed automatically on the processors
     Other Variables can be initialised using further arguments (e.g. v=... for a Variable named 'v')
     """
 
-    def __init__(self, fieldset, pclass=JITParticle, lon=None, lat=None, depth=None, time=None, repeatdt=None, lonlatdepth_dtype=None, **kwargs):
+    def __init__(self, fieldset, pclass=JITParticle, lon=None, lat=None, depth=None, time=None, repeatdt=None, lonlatdepth_dtype=None, pid_orig=None, **kwargs):
         self.fieldset = fieldset
         self.fieldset.check_complete()
+        partitions = kwargs.pop('partitions', None)
 
-        def convert_to_list(var):
-            # Convert numpy arrays and single integers/floats to one-dimensional lists
-            if isinstance(var, (int, float)):
-                return [var]
-            elif isinstance(var, np.ndarray):
+        def convert_to_array(var):
+            # Convert lists and single integers/floats to one-dimensional numpy arrays
+            if isinstance(var, np.ndarray):
                 return var.flatten()
-            return var
-        if lon is None:
-            lon = []
-        if lat is None:
-            lat = []
+            elif isinstance(var, (int, float, np.float32, np.int32)):
+                return np.array([var])
+            else:
+                return np.array(var)
+
+        lon = np.empty(shape=0) if lon is None else convert_to_array(lon)
+        lat = np.empty(shape=0) if lat is None else convert_to_array(lat)
+        if pid_orig is None:
+            pid_orig = np.arange(lon.size)
+        pid = pid_orig + pclass.lastID
 
-        lon = convert_to_list(lon)
-        lat = convert_to_list(lat)
         if depth is None:
             mindepth, _ = self.fieldset.gridset.dimrange('depth')
-            depth = np.ones(len(lon)) * mindepth
-        depth = convert_to_list(depth)
-        assert len(lon) == len(lat) and len(lon) == len(depth), (
+            depth = np.ones(lon.size) * mindepth
+        else:
+            depth = convert_to_array(depth)
+        assert lon.size == lat.size and lon.size == depth.size, (
             'lon, lat, depth don''t all have the same lenghts')
 
-        time = time.tolist() if isinstance(time, np.ndarray) else time
-        time = [time] * len(lat) if not isinstance(time, list) else time
-        time = [np.datetime64(t) if isinstance(t, datetime) else t for t in time]
-        time = [np.datetime64(t) if isinstance(t, date) else t for t in time]
+        time = convert_to_array(time)
+        time = np.repeat(time, lon.size) if time.size == 1 else time
+        if time.size > 0 and type(time[0]) in [datetime, date]:
+            time = np.array([np.datetime64(t) for t in time])
         self.time_origin = fieldset.time_origin
-        if len(time) > 0 and isinstance(time[0], np.timedelta64) and not self.time_origin:
+        if time.size > 0 and isinstance(time[0], np.timedelta64) and not self.time_origin:
             raise NotImplementedError('If fieldset.time_origin is not a date, time of a particle must be a double')
-        time = [self.time_origin.reltime(t) if isinstance(t, np.datetime64) else t for t in time]
+        time = np.array([self.time_origin.reltime(t) if isinstance(t, np.datetime64) else t for t in time])
+        assert lon.size == time.size, (
+            'time and positions (lon, lat, depth) don''t have the same lengths.')
+
+        if partitions is not None and partitions is not False:
+            partitions = convert_to_array(partitions)
 
         for kwvar in kwargs:
-            kwargs[kwvar] = convert_to_list(kwargs[kwvar])
+            kwargs[kwvar] = convert_to_array(kwargs[kwvar])
+            assert lon.size == kwargs[kwvar].size, (
+                '%s and positions (lon, lat, depth) don''t have the same lengths.' % kwargs[kwvar])
 
-        assert len(lon) == len(time), (
-            'time and positions (lon, lat, depth) don''t have the same '
-            'lengths.')
+        offset = np.max(pid) if len(pid) > 0 else -1
+        if MPI:
+            mpi_comm = MPI.COMM_WORLD
+            mpi_rank = mpi_comm.Get_rank()
+            mpi_size = mpi_comm.Get_size()
+
+            if lon.size < mpi_size and mpi_size > 1:
+                raise RuntimeError('Cannot initialise with fewer particles than MPI processors')
+
+            if mpi_size > 1:
+                if partitions is not False:
+                    if partitions is None:
+                        if mpi_rank == 0:
+                            coords = np.vstack((lon, lat)).transpose()
+                            kmeans = KMeans(n_clusters=mpi_size, random_state=0).fit(coords)
+                            partitions = kmeans.labels_
+                        else:
+                            partitions = None
+                        partitions = mpi_comm.bcast(partitions, root=0)
+                    elif np.max(partitions >= mpi_rank):
+                        raise RuntimeError('Particle partitions must vary between 0 and the number of mpi procs')
+                    lon = lon[partitions == mpi_rank]
+                    lat = lat[partitions == mpi_rank]
+                    time = time[partitions == mpi_rank]
+                    depth = depth[partitions == mpi_rank]
+                    pid = pid[partitions == mpi_rank]
+                    for kwvar in kwargs:
+                        kwargs[kwvar] = kwargs[kwvar][partitions == mpi_rank]
+                offset = MPI.COMM_WORLD.allreduce(offset, op=MPI.MAX)
 
         self.repeatdt = repeatdt.total_seconds() if isinstance(repeatdt, delta) else repeatdt
         if self.repeatdt:
@@ -85,9 +132,13 @@ class ParticleSet(object):
             self.repeat_starttime = time[0]
             self.repeatlon = lon
             self.repeatlat = lat
+            if not hasattr(self, 'repeatpid'):
+                self.repeatpid = pid - pclass.lastID
             self.repeatdepth = depth
             self.repeatpclass = pclass
+            self.partitions = partitions
             self.repeatkwargs = kwargs
+        pclass.setLastID(offset+1)
 
         if lonlatdepth_dtype is None:
             self.lonlatdepth_dtype = self.lonlatdepth_dtype_from_field_interp_method(fieldset.U)
@@ -97,14 +148,13 @@ class ParticleSet(object):
             'lon lat depth precision should be set to either np.float32 or np.float64'
         JITParticle.set_lonlatdepth_dtype(self.lonlatdepth_dtype)
 
-        size = len(lon)
-        self.particles = np.empty(size, dtype=pclass)
+        self.particles = np.empty(lon.size, dtype=pclass)
         self.ptype = pclass.getPType()
         self.kernel = None
 
         if self.ptype.uses_jit:
             # Allocate underlying data for C-allocated particles
-            self._particle_data = np.empty(size, dtype=self.ptype.dtype)
+            self._particle_data = np.empty(lon.size, dtype=self.ptype.dtype)
 
             def cptr(i):
                 return self._particle_data[i]
@@ -113,12 +163,12 @@ class ParticleSet(object):
                 return None
 
         if lon is not None and lat is not None:
-            # Initialise from lists of lon/lat coordinates
-            assert size == len(lon) and size == len(lat), (
+            # Initialise from arrays of lon/lat coordinates
+            assert self.particles.size == lon.size and self.particles.size == lat.size, (
                 'Size of ParticleSet does not match lenght of lon and lat.')
 
-            for i in range(size):
-                self.particles[i] = pclass(lon[i], lat[i], fieldset=fieldset, depth=depth[i], cptr=cptr(i), time=time[i])
+            for i in range(lon.size):
+                self.particles[i] = pclass(lon[i], lat[i], pid[i], fieldset=fieldset, depth=depth[i], cptr=cptr(i), time=time[i])
                 # Set other Variables if provided
                 for kwvar in kwargs:
                     if not hasattr(self.particles[i], kwvar):
@@ -191,10 +241,11 @@ class ParticleSet(object):
         """
 
         if mode == 'monte_carlo':
+            data = start_field.data if isinstance(start_field.data, np.ndarray) else np.array(start_field.data)
             if start_field.interp_method == 'cgrid_tracer':
-                p_interior = np.squeeze(start_field.data[0, 1:, 1:])
+                p_interior = np.squeeze(data[0, 1:, 1:])
             else:  # if A-grid
-                d = start_field.data
+                d = data
                 p_interior = (d[0, :-1, :-1] + d[0, 1:, :-1] + d[0, :-1, 1:] + d[0, 1:, 1:])/4.
                 p_interior = np.where(d[0, :-1, :-1] == 0, 0, p_interior)
                 p_interior = np.where(d[0, 1:, :-1] == 0, 0, p_interior)
@@ -262,8 +313,8 @@ class ParticleSet(object):
         """Method to add particles to the ParticleSet"""
         if isinstance(particles, ParticleSet):
             particles = particles.particles
-        if not isinstance(particles, collections.Iterable):
-            particles = [particles]
+        else:
+            raise NotImplementedError('Only ParticleSets can be added to a ParticleSet')
         self.particles = np.append(self.particles, particles)
         if self.ptype.uses_jit:
             particles_data = [p._cptr for p in particles]
@@ -411,6 +462,10 @@ class ParticleSet(object):
         while (time < endtime and dt > 0) or (time > endtime and dt < 0) or dt == 0:
             if verbose_progress is None and time_module.time() - walltime_start > 10:
                 # Showing progressbar if runtime > 10 seconds
+                if output_file:
+                    logger.info('Temporary output files are stored in %s.' % output_file.tempwritedir_base)
+                    logger.info('You can use "parcels_convert_npydir_to_netcdf %s" to convert these '
+                                'to a NetCDF file during the run.' % output_file.tempwritedir_base)
                 pbar = progressbar.ProgressBar(max_value=abs(endtime - _starttime)).start()
                 verbose_progress = True
             if dt > 0:
@@ -422,7 +477,7 @@ class ParticleSet(object):
                 pset_new = ParticleSet(fieldset=self.fieldset, time=time, lon=self.repeatlon,
                                        lat=self.repeatlat, depth=self.repeatdepth,
                                        pclass=self.repeatpclass, lonlatdepth_dtype=self.lonlatdepth_dtype,
-                                       **self.repeatkwargs)
+                                       partitions=False, pid_orig=self.repeatpid, **self.repeatkwargs)
                 for p in pset_new:
                     p.dt = dt
                 self.add(pset_new)
@@ -434,7 +489,8 @@ class ParticleSet(object):
             if abs(time-next_movie) < tol:
                 self.show(field=movie_background_field, show_time=time, animation=True)
                 next_movie += moviedt * np.sign(dt)
-            next_input = self.fieldset.computeTimeChunk(time, dt)
+            if time != endtime:
+                next_input = self.fieldset.computeTimeChunk(time, dt)
             if dt == 0:
                 break
             if verbose_progress:
